@@ -153,13 +153,12 @@ data KnownExtension =
   -- | [GHC &#xa7; 7.6.3.5] Allow type signatures in instances.
   | InstanceSigs
 
-  -- | [GHC &#xa7; 7.3.8] Allows recursive bindings in @do@ blocks,
-  -- using the @rec@ keyword.
+  -- | [GHC &#xa7; 7.3.8] Deprecated in GHC.  Allows recursive
+  -- bindings in @do@ blocks, using the @rec@ keyword.
   | DoRec
 
-  -- | [GHC &#xa7; 7.3.8.2] Deprecated in GHC.  Allows recursive bindings
-  -- using @mdo@, a variant of @do@.  @DoRec@ provides a different,
-  -- preferred syntax.
+  -- | [GHC &#xa7; 7.3.8.2] Allows recursive bindings using @mdo@, a
+  -- variant of @do@, and @rec@.
   | RecursiveDo
 
   -- | [GHC &#xa7; 7.3.9] Provide syntax for writing list
@@ -270,6 +269,9 @@ data KnownExtension =
   -- @deriving@ for any class with an instance for the underlying type.
   | GeneralizedNewtypeDeriving
 
+  -- | [GHC &#xa7; 9.6.10] Allow use of any typeclass in deriving clauses.
+  | DeriveAnyClass
+
   -- | [Hugs &#xa7; 7.1] Enable the \"Trex\" extensible records system.
   | ExtensibleRecords
 
@@ -375,6 +377,9 @@ data KnownExtension =
 
   | LambdaCase
 
+  -- | [GHC &#xa7; 7.3.20] Allow case expressions with no alternatives.
+  | EmptyCase
+
   -- | [GHC &#xa7; 7.8.6] Deprecated in GHC 6.12 and will be removed in
   -- GHC 7.  Allow a type variable to be instantiated at a
   -- polymorphic type.
@@ -465,6 +470,10 @@ data KnownExtension =
   -- foreign function interface.
   | CApiFFI
 
+  -- | GHCJS FFI extension with convenient import patterns,
+  -- asynchronous FFI and a JSVal FFI type
+  | JavaScriptFFI
+
   -- | [GHC &#xa7; 7.3.27] Enable explicit namespaces in import/export.
   | ExplicitNamespaces
 
@@ -530,6 +539,13 @@ data KnownExtension =
   | TypeFamilyDependencies
 
   | OverloadedLabels
+
+  -- | Allow multiple @deriving@ clauses, each optionally qualified with a
+  -- /strategy/.
+  | DerivingStrategies
+
+  | UnboxedSums
+
   deriving (Show, Read, Eq, Ord, Enum, Bounded, Data, Typeable)
 
 -- | Certain extensions imply other extensions, and this function
@@ -713,4 +729,3 @@ toExtensionList lang exts' =
           go acc (EnableExtension  x : exts) = go (x : acc)           exts
           -- We just throw away UnknownExtensions
           go acc (_ : exts) = go acc exts
-
